/* CTP protocol logic layer
 *
 * Copyright (C) 2010 Beijing Greatal Technologies Co. Ltd.
 *
 * This layer implements CTP protocol details, and located between HW and
 * APP layers. This file is independent on HW platform and could be reused
 * on your HW without any change.
 *
 */
#include "ctp.h"

#if 1
#include <stdio.h>
#define ctp_dbg printf
#else
void ctp_dbg(char *fmt, ...)
{
        return;
}
#endif
/********************************************************
 * DO NOT change the codes in this file
 * This C file does NOT depend on any libary or OS, can be
 * complied on various platform: C51, ARM, DSP, PC etc.
 *******************************************************/
/*Global control status count for debug*/
u32 ok_count = 0;
u32 fail_count = 0;
u32 unknown_count = 0;
u32 error_count = 0;

/********************************************
 * Libary implementation
 *******************************************/
/*
 * String functions
 */
u32 str_len(const s8 *str)
{
    const s8 *s;
    for (s = str; *s; ++s);
    return(s - str);
}

s8 *str_cpy(s8 *dest, const s8 *src)
{
  u32 i;
  for (i=0; src[i] != '\0'; ++i)
    dest[i] = src[i];
  dest[i] = '\0';
  return dest;
}

s8 *str_cat(s8 *dest, const s8 *src)
{
    str_cpy(dest + str_len(dest), src);
    return dest;
}

/*
 * FIFO functions
 */
#define WR 1
#define RD 0
#define FIFO_SZ	64

typedef struct fifo{
	s8 buffer[FIFO_SZ];	/* Circular buffer */
	u32 w_idx;	/* Index of write */
	u32 r_idx;	/* Index of read */
}fifo_t;

/* return 0: write or read OKAY
         <0: unreadable
*/ 
static s32 fifo_ctl(fifo_t *p, s8 *value, const u8 flag)
{
	s32 ret = 0;
	static u32 widx;

	widx = p->w_idx;    /* Save the old index before proceeding */
	if (flag == WR) {
        	if (((p->w_idx + 1) % FIFO_SZ) == p->r_idx) {
		/* Adjust read index since buffer is full */
		/* Keep the latest one and drop the oldest one */
			p->r_idx = (p->r_idx + 1) % FIFO_SZ;
			ctp_dbg("Error: Fifo is full!\n");
		}
		p->buffer[p->w_idx] = *value;
        	p->w_idx = (p->w_idx + 1) % FIFO_SZ;
	}
	else {
		if ((p->r_idx == widx)) {
			ret = -1;
		}	
		else {
			*value = p->buffer[p->r_idx];
			p->r_idx = (p->r_idx + 1) % FIFO_SZ;
		}
	}

	return ret;
}

static void fifo_init(fifo_t *p)
{
    p->r_idx = 0;
    p->w_idx = 0;
}

/***************************************************************
 * CTP protocol Implementation
 ***************************************************************/
#define	CTP_BUF_SZ	64
#define MSG_MAX_COUNT	10
#define MSGQ_EMPTY	-1
#define MSGQ_MATCH	0
#define MSGQ_NONMATCH	1
#define MSGQ_FAILURE	2

typedef struct msg_buf{
	s8 type;	/* E: event, C: control Q: query */
	s8 result;	/* +: okay, -: failure ?: unkown obj, !: format error */
	s8 content[CTP_BUF_SZ];	/* message content */
}msg_t;

static msg_t msg[MSG_MAX_COUNT];
static fifo_t ctp_fifo;
static s8 ctp_buf[CTP_BUF_SZ];

/*
 * static functions
 */
static void msg2npv(s8 *s, s8 *name, s8 *prop, s8 *value)
{
	s8 *p = s;
	while (1) {
		*name = (*p == '.' ? '\0' : *p);
		p++;
		if (*name == '\0')
			break;
		name++;
	}

	while (1) {
		*prop = (*p == '=' ? '\0' : *p);
		p++;
		if (*prop == '\0')
			break;
		prop++;
	}
	str_cpy(value, p);
}

static s32 msg2v(s8 *s, s8 *name, s8 *prop, s8 *value)
{
	s8 *p = s;
	s32 ret = -1;
	s8 tmp;

	while (1) {
		tmp = (*p == '.' ? '\0' : *p);
		p++;
		if (*name != tmp)
			goto end;
		if (*name == '\0')
			break;
		name++;
	}

	while (1) {
		tmp = (*p == '=' ? '\0' : *p);
		p++;
		if (*prop != tmp)
			goto end;
		if (*prop == '\0')
			break;
		prop++;
	}
	ret = 0;
	str_cpy(value, p);
end:
	return ret;
}

/* return 0: no ctp frame available;
        >=0: ctp frame available, return frame size  (not including '\0')
         -1: error
 */
static s32 read_ctp_frame(void)
{
	s32 ret = 0;
	static s8 *p = ctp_buf;

	while(1) {
		if (fifo_ctl(&ctp_fifo, p, RD))
			goto empty;
		else if (*p == '\r' || *p == '\n') {
			*p = '\0';
			ret = p - ctp_buf;
			goto reset;
		}

		if (p++ >= (ctp_buf + CTP_BUF_SZ)) {
			ctp_dbg("Error: ctp_buf overflow! reset buf\n");
			ret = -1;
			goto reset;
		}
	}
reset:
	p = ctp_buf;
empty:
	return ret;
}

void ctp_init(void)
{
	hw_init();
	fifo_init(&ctp_fifo);
}

/* return:
 0: matched
-1: empty
 1: not matched
*/
static s32 ctp_core(u8 type, s8 *name, s8 *prop, s8 *value)
{
	s8 tmp[64];
	s32 ret = MSGQ_NONMATCH;
	s32 sz, i;

	/* update msg queue from fifo*/
	if (read_ctp_frame() > 0) {
		if (ctp_buf[0] == 'C') {
			if (ctp_buf[1] == '+')
				ok_count++;
			else if (ctp_buf[1] == '-')
				fail_count++;
			else if (ctp_buf[1] == '?')
				unknown_count++;
			else
				error_count++;
		}

		if (ctp_buf[0] == 'E' || ctp_buf[0] == 'Q') {
			for (i = 0; i < MSG_MAX_COUNT; i++) {
				if (msg[i].type == 0) {
					str_cpy((s8 *)&msg[i], ctp_buf);
					ctp_dbg("->%s\n", ctp_buf);
					break;
				}
			}
			if (i >= MSG_MAX_COUNT)
				ctp_dbg("Error:msg queue overrun!\n");
		}
	}

	/* read wanted msg */
	for (i = 0; i < MSG_MAX_COUNT; i++) {
		if (msg[i].type == type) {
			if (msg[i].result == '+' && type == 'E') {
				msg2npv(msg[i].content, name, prop, value);
				msg[i].type = 0;
				ret = MSGQ_MATCH;
				goto end;
			}
			else if (type == 'Q') {
				if (!msg2v(msg[i].content, name, prop, value)) {
					msg[i].type = 0;
					ret = (msg[i].result == '+' ? MSGQ_MATCH : MSGQ_FAILURE);
					goto end;
				}
				else
					ctp_dbg("Warning: orphan query exists!\n");
			}
		}
	}

	/* wait and copy hw data to fifo */
	sz = hw_read(tmp, 64);
	if (sz > 0) {
		for (i = 0; i < sz; i++)
			fifo_ctl(&ctp_fifo, &tmp[i], WR);
	}
end:
	return ret;
}

/*
 * Exported functions
 */
s32 ctp_control(s8 *name, s8 *prop, s8 *value)
{
	hw_write(name, str_len(name));
	hw_write(".", 1);
	hw_write(prop, str_len(prop));
	hw_write("=", 1);
	hw_write(value, str_len(value));
	hw_write("\r", 1);
	ctp_dbg("<-%s.%s=%s\n", name, prop, value);
	return 0;
}

/* 
 * block until getting query result from cooky whatever
 * hw_read() get blocked.
 * return 0: Okay, store result in the argument "value"
 *       -1: Failure 
 */
s32 ctp_query(s8 *name, s8 *prop, s8 *value)
{
	s32 ret;

	hw_write(name, str_len(name));
	hw_write(".", 1);
	hw_write(prop, str_len(prop));
	hw_write("?", 1);
	hw_write("\r", 1);
	ctp_dbg("<-%s.%s?\n", name, prop);

	while (1) {
		ret = ctp_core('Q', name, prop, value);
		if (ret == MSGQ_MATCH || ret == MSGQ_FAILURE)
			break;
	}

	return (ret == MSGQ_MATCH ? 0 : -1);
}

/* 
 * block in this call if hw_read() get blocked.
 * return 0: Okay, get a event
 */
s32 ctp_event(s8 *name, s8 *prop, s8 *value)
{
	s32 ret;

	ret = ctp_core('E', name, prop, value);

	return (ret == MSGQ_MATCH) ? 0 : -1;
}
